/*
  lessutf.c - An input pipe for less to display UTF files.

  Jason Hood, 11 to 16 September, 2014.

  Less sees UTF-16 files as binary.  I had a go at reading them directly, but
  it didn't quite work, hence the filter.  If a BOM is not present it looks at
  the contents to see if it's UTF-16 (a significant number of ASCII values, or
  a sufficient number of spaces), or works as UTF-8 (no lead or trail bytes out
  of place).

  3 & 4 November, 2014:
  - use the current code page;
  * use bold "?" for unrepresentables, not "^_".

  28 November, 2014:
  - can't test little-endian + big-endian is whole file, dummy!

  20 December, 2014:
  + read standard input;
  * increase the buffer size (32768 bytes instead of 8192);
  * don't convert to OEM - less will recognise this filter and use
    WriteConsoleW, eliminating the need to use UTF-8 mode.
*/

#define _CRT_SECURE_NO_WARNINGS
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <fcntl.h>
#include <share.h>

#define UTF8	1
#define UTF16LE 2
#define UTF16BE 3


// UTF-8 routines copied from LESS.

#define IS_UTF8_TRAIL(c)    (((c) & 0xC0) == 0x80)
#define IS_UTF8_INVALID(c)  (((c) & 0xFE) == 0xFE)
#define IS_UTF8_LEAD(c)     (((c) & 0xC0) == 0xC0 && !IS_UTF8_INVALID(c))

// Get the length of a UTF-8 character in bytes.
int utf_len( BYTE ch )
{
  if ((ch & 0x80) == 0)    return 1;
  if ((ch & 0xE0) == 0xC0) return 2;
  if ((ch & 0xF0) == 0xE0) return 3;
  if ((ch & 0xF8) == 0xF0) return 4;
  if ((ch & 0xFC) == 0xF8) return 5;
  if ((ch & 0xFE) == 0xFC) return 6;

  // Invalid UTF-8 encoding.
  return 1;
}

// Does the parameter point to the lead byte of a well-formed UTF-8 character?
BOOL is_utf8_well_formed( PBYTE s )
{
  int i;
  int len;

  if (IS_UTF8_INVALID( s[0] ))
    return FALSE;

  len = utf_len( s[0] );
  if (len == 1)
    return TRUE;

  if (len == 2)
  {
    if (s[0] < 0xC2)
      return FALSE;
  }
  else
  {
    BYTE mask;
    mask = (~((1 << (8-len)) - 1)) & 0xFF;
    if (s[0] == mask && (s[1] & mask) == 0x80)
      return FALSE;
  }

  for (i = 1; i < len; ++i)
    if (!IS_UTF8_TRAIL( s[i] ))
      return FALSE;

  return TRUE;
}


int main( int argc, char* argv[] )
{
  int	fd;
  #define BUF_SIZE 32768
  BYTE	buf[BUF_SIZE], buf8[BUF_SIZE*2];
  int	size, len, skipped, n;
  int	utf, bom;
  int	utf8;
  char* charset;

  if (argc != 2)
    return 0;

  if (argv[1][0] == '-' && argv[1][1] == '\0')
  {
    fd = 0;
    _setmode( 0, _O_BINARY );
  }
  else
  {
    fd = _sopen( argv[1], _O_RDONLY | _O_BINARY | _O_SEQUENTIAL, _SH_DENYNO );
    if (fd < 0)
      return 0;

    size = _lseek( fd, 0, SEEK_END );
    _lseek( fd, 0, SEEK_SET );
  }
  _setmode( 1, _O_BINARY );

  utf8 = (GetConsoleOutputCP() == CP_UTF8);
  charset = getenv("LESSCHARSET");
  if (charset && strcmp( charset, "utf-8" ) == 0)
    utf8 |= 2;

  skipped = 0;
  utf = 0;
  bom = 0;
  len = _read( fd, buf, BUF_SIZE );
  if (len > 3 && buf[0] == 0xEF && buf[1] == 0xBB && buf[2] == 0xBF)
  {
    bom = 3;
    utf = UTF8;
  }
  else
  {
    if (fd == 0)
      size = len;
    if (size > 2 && !(size & 1))
    {
      if (buf[0] == 0xFF && buf[1] == 0xFE)
      {
	bom = 2;
	utf = UTF16LE;
      }
      else if (buf[0] == 0xFE && buf[1] == 0xFF)
      {
	bom = 2;
	utf = UTF16BE;
      }
      else
      {
	int le_count = 0, be_count = 0;
	int le_space = 0, be_space = 0;
	for (n = 0; n < len; n += 2)
	{
	  if (buf[n] == 0)
	  {
	    ++be_count;
	    if (buf[n+1] == ' ')
	      ++be_space;
	  }
	  if (buf[n+1] == 0)
	  {
	    ++le_count;
	    if (buf[n] == ' ')
	      ++le_space;
	  }
	}
	if (le_count > len/2 / 2)
	  utf = UTF16LE;
	else if (be_count > len/2 / 2)
	  utf = UTF16BE;
	else if (le_space > len/2 / 16)
	  utf = UTF16LE;
	else if (be_space > len/2 / 16)
	  utf = UTF16BE;
      }
    }
  }
  if (!utf && !utf8)
  {
    // Didn't find a BOM or detect UTF-16, see if it looks like UTF-8.
    BOOL lead = FALSE;

    // Skip a possibly truncated sequence.
    while (IS_UTF8_TRAIL( buf[len-1] ) && len > 0)
      --len, ++skipped;
    if (len > 0 && IS_UTF8_LEAD( buf[len-1] ))
      --len, ++skipped;

    for (n = 0; n < len; n += size)
    {
      if (buf[n] == 0 || !is_utf8_well_formed( buf + n ))
	break;
      size = utf_len( buf[n] );
      if (size != 1)
	lead = TRUE;
    }
    if (lead && n == len && len != 0)
      utf = UTF8;
  }
  if (fd != 0 && (!utf || (utf == UTF8 && (utf8 & 2))))
  {
    _close( fd );
    return 0;
  }

  if (bom)
  {
    len -= bom;
    memcpy( buf, buf + bom, len );
  }

  // Still have to copy standard input, let less know if it's UTF-8.
  if (fd == 0)
  {
    _write( 1, &utf, 1 );
    if (!utf)
      utf = UTF8;
  }

  if (utf == UTF8)
  {
    len += skipped;
    do
    {
      _write( 1, buf, len );
      len = _read( fd, buf, sizeof(buf) );
    } while (len > 0);
  }
  else
  {
    do
    {
      // Translate big-endian to little.
      if (utf == UTF16BE)
      {
	for (n = 0; n < len; n += 2)
	{
	  BYTE t = buf[n];
	  buf[n] = buf[n+1];
	  buf[n+1] = t;
	}
      }

      // Avoid splitting a surrogate pair.
      if (buf[len-1] >= 0xD8 && buf[len-1] < 0xDC && len > 2)
      {
	len -= 2;
	skipped = 2;
      }

      len = WideCharToMultiByte( CP_UTF8, 0, (LPCWSTR)buf, len / 2,
				 buf8, sizeof(buf8), NULL, NULL );
      _write( 1, buf8, len );

      if (skipped)
      {
	buf[0] = buf[len];
	buf[1] = buf[len+1];
      }
      len = _read( fd, buf + skipped, BUF_SIZE - skipped );
      skipped = 0;
    } while (len > 0);
  }

  if (fd != 0)
    _close( fd );
  return 0;
}
